/**************************************************************
* COS265 Data Structures & Algorithms                         *
* Project 4 - Pathfinding                                     *
* by Chad Jordan 11/05/07                                     *
*                                                             *
* Java implementation that allows the user to interactively   *
* edit the terrain heightmap and communicate with the         *
* Pathfinder Visualizer. The methodologies in this program    *
* allow editing operations such as smoothing, increasing/     *
* decreasing height, and generating a factal terrain.         *
* Dependencies: Terrain.java, PathfinderVisualizer.java       *  
**************************************************************/
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.awt.geom.Point2D;
import java.util.Random;

public class TerrainEditor {

    private Terrain terrain;
    private PathfinderVisualizer visualizer;

    public TerrainEditor(Terrain terrain, PathfinderVisualizer visualizer) {
        this.terrain = terrain;
        this.visualizer = visualizer;
    }

    public void smoothTerrain() {
        int[][] smoothedHeightMap = new int[terrain.getHeightMap().length][terrain.getHeightMap()[0].length];
        for (int i = 0; i < terrain.getHeightMap().length; i++) {
            for (int j = 0; j < terrain.getHeightMap()[0].length; j++) {
                int sum = terrain.getHeightMap()[i][j];
                int count = 1;
                if (i > 0) {
                    sum += terrain.getHeightMap()[i - 1][j];
                    count++;
                }
                if (i < terrain.getHeightMap().length - 1) {
                    sum += terrain.getHeightMap()[i + 1][j];
                    count++;
                }
                if (j > 0) {
                    sum += terrain.getHeightMap()[i][j - 1];
                    count++;
                }
                if (j < terrain.getHeightMap()[0].length - 1) {
                    sum += terrain.getHeightMap()[i][j + 1];
                    count++;
                }
                smoothedHeightMap[i][j] = sum / count;
            }
        }
        terrain.setHeightMap(smoothedHeightMap);
        visualizer.repaint();
    }

    public void increaseHeight() {
        for (int i = 0; i < terrain.getHeightMap().length; i++) {
            for (int j = 0; j < terrain.getHeightMap()[0].length; j++) {
                terrain.getHeightMap()[i][j]++;
            }
        }
        visualizer.repaint();
    }

    public void decreaseHeight() {
        for (int i = 0; i < terrain.getHeightMap().length; i++) {
            for (int j = 0; j < terrain.getHeightMap()[0].length; j++) {
                terrain.getHeightMap()[i][j]--;
            }
        }
        visualizer.repaint();
    }

    public void generateFractalTerrain() {
        int size = terrain.getHeightMap().length;
        int[][] fractalHeightMap = new int[size][size];
        Random random = new Random();
        for (int length = size - 1; length >= 2; length /= 2) {
            int half = length / 2;
            double roughness = 0.5;
            for (int x = 0; x < size - 1; x += length) {
                for (int y = 0; y < size - 1; y += length) {
                    double avg = (fractalHeightMap[x][y] + fractalHeightMap[x + length][y] +
                                  fractalHeightMap[x][y + length] + fractalHeightMap[x + length][y + length]) / 4.0;
                    fractalHeightMap[x + half][y + half] = (int) (avg + (random.nextDouble() * 2 - 1) * roughness * length);
                }
            }
            for (int x = 0; x < size - 1; x += half) {
                for (int y = (x + half) % length; y < size - 1; y += length) {
                    double avg = (fractalHeightMap[(x - half + size) % size][y] + fractalHeightMap[(x + half) % size][y] +
                                  fractalHeightMap[x][(y + half) % size] + fractalHeightMap[x][(y - half + size) % size]) / 4.0;
                    fractalHeightMap[x][y] = (int) (avg + (random.nextDouble() * 2 - 1) * roughness * length);
                }
            }
            roughness /= 2.0;
        }
        terrain.setHeightMap(fractalHeightMap);
        visualizer.repaint();
    }
    // Default heightMap size 50x50
    public static void main(String[] args) {
        int[][] heightMap = new int[50][50]; 
        Terrain terrain = new Terrain(heightMap);

        Pathfinder pathfinder = new AStarPathfinder();
        Walker walker = new Walker(new Point2D.Double(0, 0), new Point2D.Double(49, 49));

        PathfinderVisualizer visualizer = new PathfinderVisualizer(terrain, pathfinder, walker);

        JFrame frame = new JFrame("Terrain Editor");
        frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        frame.getContentPane().add(visualizer);
        frame.pack();
        frame.setLocationRelativeTo(null);
        frame.setVisible(true);

        TerrainEditor editor = new TerrainEditor(terrain, visualizer);

        JMenuBar menuBar = new JMenuBar();
        JMenu menu = new JMenu("Terrain");
        menuBar.add(menu);

        JMenuItem smoothItem = new JMenuItem("Smooth Terrain");
        smoothItem.addActionListener(e -> editor.smoothTerrain());
        menu.add(smoothItem);

        JMenuItem increaseItem = new JMenuItem("Increase Height");
        increaseItem.addActionListener(e -> editor.increaseHeight());
        menu.add(increaseItem);

        JMenuItem decreaseItem = new JMenuItem("Decrease Height");
        decreaseItem.addActionListener(e -> editor.decreaseHeight());
        menu.add(decreaseItem);

        JMenuItem fractalItem = new JMenuItem("Generate Fractal Terrain");
        fractalItem.addActionListener(e -> editor.generateFractalTerrain());
        menu.add(fractalItem);

        frame.setJMenuBar(menuBar);
    }
}